-- By Verne Tice (vstice@gmail.com)
-- Inspired by cputime.e by Daniel Kluss(xerox_irs@lvcm.com)
--   and, I admit, copies some code from cputime.e
-- Must use pentium class or above to read the time stamp of the cpu.
-- 
-- Please post comments, questions to Euforum
--
-- Also, if you use hrtime.e for timing please post any results
-- you may feel might be of interest to Euforum.
--
-- Cautions:
--  If used with a CPU which can change its clock speed dynamically,
--  generally to save power, use of this program will give VERY 
--  questionable results unless the computer is configured to not
--  do such speed changes.

--  Most computers which support such dynamic speed changes give the user
--  the option of disabling this via the BIOS.  Refer to your computer's
--  manual for instructions.

without warning
include machine.e
--include asm.e  -- VST no longer needed
include misc.e

atom start_cpu_time,start_clock_time,clock_mult
atom temp -- VST
atom cpu_speed

-- ##################################################################
--
global constant HLT = allocate(2) -- use in windows dos box only, crashs in plain dos
--
-- ##################################################################

-- call(HLT)
--  gives control back to OS for 0 or a few ticks (timeslices)
--  crashes in straight DOS, but works fine in windows DOS box,
--  windows console, win32 app, and should work OK in Linux or FreeBSD


poke(HLT, {
    #F4,                    --    0: hlt
    #C3})                   --    1: ret

-- eliminated need for including asm.e by using it to "compile"
-- asm code, then copied "compiled" code into cputime.e
-- Also modified code to save and restore registers, and to serialize
-- cpu per Intel recc before doing rdtsc instruction.

-- By serialize, I mean that all instructions currently in pipeline
--  are completed before the rdtsc instruction is executed.
--  Although this can cost additional clock cycles between calls,
--  measurements should be more repeatible.
-- Of course, if the amount of time between calls is much greater than
--  the time for 1 clock cycle, then this concern becomes moot.

constant  CPU_MEM=allocate(8)

--constant
--    Read_Time_Stamp_Asm = get_asm(
--        "pusha "&           -- Save registers to stack
--        "cpuid "&           -- Serialize per Intel recc    
--        "rdtsc "&           -- Read TimeStampCounter ( 64-bit )
--        "mov [cm1],eax "&   -- Save low-order 32-bits into CPU_MEM
--        "mov [cm2],edx "&   -- Save high-order 32-bits into CPU_MEM+4
--        "popa "&            -- Restore registers
--        "ret " )            -- Return to calling program
--    resolve_param("cm1",CPU_MEM)
--    resolve_param("cm2",CPU_MEM+4)

constant Read_Time_Stamp_Asm = allocate(18)
poke(Read_Time_Stamp_Asm, {
    #60,                    --    0: pusha
    #0F,#A2,                --    1: cpuid
    #0F,#31,                --    3: rdtsc
    #A3,#00,#00,#00,#00,    --    5: mov [cm1],eax (6)
    #89,#15,#00,#00,#00,#00,--    A: mov [cm2],edx (12)
    #61,                    --   10: popa
    #C3})                   --   11: ret
poke4(Read_Time_Stamp_Asm + 6, CPU_MEM)
poke4(Read_Time_Stamp_Asm + 12, CPU_MEM + 4)

-- ##################################################################
--
global function cpu_time()--get the clock count, most accurate timer on the computer
--
-- ##################################################################

    call(Read_Time_Stamp_Asm)
    return peek4u(CPU_MEM)+peek4u(CPU_MEM+4)*#100000000
end function

-- ##################################################################
--
global function get_cpu_speed()
--
-- ##################################################################

    return cpu_speed/1.0e6
end function

-- For cases where there is too much background activity for
--  reliable measurement of cpu speed.

-- ##################################################################
--
global function set_cpu_speed(atom mhz)
--
-- ##################################################################


-- Allow user to set cpu speed (in MHz)
-- Useful for cases where measured cpu speed shows too much variation
-- Returns previous (measured?) cpu speed (in MHz)

    atom old_spd
    old_spd = cpu_speed
    cpu_speed = mhz * 1.0e6
    clock_mult=1.0/cpu_speed
    return old_spd/1.0e6

end function

-- changed inline code to function to allow user to call
-- useful for evaluation of reliability of cpu speed measurement

-- ##################################################################
--
global function measure_cpu_speed(integer ticks)
--
-- ##################################################################


-- Argument
--  ticks - measurement duration (10 to 20 works well)
--          
--  For DOS32 temporarily sets tick_rate to 279

-- New (sync sleep and time to clock tick)
-- Waits for "tick" before starting, thereby eliminating variance due
--      to starting at an unknown point in "tick" duration
-- Assumes that the time to do assignments and read TimeStampCounter
--      is much less than 1 tick

-- NOTE: resets base time to zero, so any pending measurements in current
--       program will be WRONG!

    if ticks <= 0 then ticks = 10 end if

    if platform()=DOS32 then
        tick_rate(997) -- set high timer resolution
    end if

	start_clock_time = time()

    -- wait for start of new tick before starting speed measurement
	while start_clock_time = time() do end while    

	start_cpu_time = cpu_time()
	start_clock_time = time()

    -- measure for specified number of ticks rather than sleep(1),
    -- or 1 whole second, which causes a noticeable delay

    for i = 1 to ticks do
        temp = time()
        while temp = time() do end while

        -- using just time() = time() in "while" caused 0 to 5 ticks to be missed.
        -- although only time between first and last tick is important,
        -- measuring for n+? number of ticks is just messy!

    end for

	cpu_speed=(cpu_time()-start_cpu_time)/(time()-start_clock_time)

    clock_mult=1.0/cpu_speed

    if platform()=DOS32 then
        tick_rate(0)--set back, cause you don't need it anymore
    end if

    return cpu_speed/1.0e6 -- prefer speed in MHz

end function

temp = measure_cpu_speed(23)        -- try to get code in cache

if platform() = WIN32 then
    temp = measure_cpu_speed(23)    -- win requires longer sample time
else
    temp = measure_cpu_speed(23)    -- dos32 997 ticks/s, linux 1000
end if

-- ##################################################################
--
global function hrtime() -- High resolution elapsed time routine
--
-- ##################################################################

    return start_clock_time+(cpu_time()-start_cpu_time)*clock_mult
end function

printf(1,"CPU Speed %8.3f Mhz\n",{get_cpu_speed()})
